@@ -18,14 +18,14 @@
 // when writing the expected output of a unit test.
 //
 // ```
-// let mut f = InternalFormatter::new();
+// let mut f = Internalexport::make_formatter();
 // let index = defmt::export::fetch_string_index();
 // foo(&mut f); // NOTE increases the interner index
-// assert_eq!(f.bytes(), &[index]);
+// assert_eq!(fetch_bytes(), [index]);
 //
-// let mut f = InternalFormatter::new();
+// let mut f = Internalexport::make_formatter();
 // foo(&mut f);
-// assert_eq!(f.bytes(), &[index.wrapping_add(1)]);
+// assert_eq!(fetch_bytes(), [index.wrapping_add(1)]);
 //                               ^^^^^^^^^^^^^^^ account for the previous `foo` call
 // ```
 //
@@ -36,221 +36,119 @@
 //
 // - the mocked index is 7 bits so its LEB128 encoding is the input byte
 
-use defmt::{
-    export::fetch_string_index, write, Debug2Format, Display2Format, Format, Formatter,
-    InternalFormatter,
-};
+use defmt::{export::fetch_string_index, write, Debug2Format, Display2Format, Format, Formatter};
 
 // Increase the 7-bit mocked interned index
-fn inc(index: u8, n: u8) -> u8 {
-    // NOTE(&) keep the highest bit at 0
-    index.wrapping_add(n) & 0x7F
+fn inc(index: u16, n: u16) -> u16 {
+    index.wrapping_add(n)
 }
 
-fn check_format_implementation(val: &(impl Format + ?Sized), expected_encoding: &[u8]) {
-    let mut f = InternalFormatter::new();
-    let g = Formatter { inner: &mut f };
-    val.format(g);
-    f.finalize();
-    assert_eq!(f.bytes(), expected_encoding);
+fn write_format<T: Format + ?Sized>(val: &T) {
+    defmt::export::istr(&T::_format_tag());
+    val._format_data();
+}
+
+macro_rules! check {
+    ([$($x:expr),* $(,)?]) => {
+        {
+            let mut v = Vec::<u8>::new();
+            $(
+                v.extend(&($x).to_le_bytes());
+            )*
+            assert_eq!(defmt::export::fetch_bytes(), v);
+        }
+    };
+}
+
+macro_rules! check_format {
+    ($format:expr, [$($x:expr),* $(,)?]  $(,)?) => {
+        {
+            let mut v = Vec::<u8>::new();
+            $(
+                v.extend(&($x).to_le_bytes());
+            )*
+            write_format($format);
+            assert_eq!(defmt::export::fetch_bytes(), v);
+        }
+    }
 }
 
 #[test]
 fn write() {
     let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
+    let g = defmt::export::make_formatter();
     write!(g, "The answer is {=u8}", 42);
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // "The answer is {=u8}",
-            42,    // u8 value
-        ]
-    );
+    check!([
+        index, // "The answer is {=u8}",
+        42u8,  // u8 value
+    ]);
 
-    let f2 = &mut InternalFormatter::new();
-    let g2 = Formatter { inner: f2 };
-    write!(g2, "The answer is {=?}", 42u8);
-    assert_eq!(
-        f2.bytes(),
-        &[
-            inc(index, 1), // "The answer is {=?}"
-            inc(index, 2), // "{=u8}" / impl Format for u8
-            42,            // u8 value
-        ]
-    );
-}
-
-#[test]
-fn booleans_max_num_bool_flags() {
-    let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
-    write!(
-        g,
-        "encode 8 bools {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool}",
-        false, true, true, false, true, false, true, true
-    );
-    assert_eq!(
-        f.bytes(),
-        &[
-            index,       // "encode 8 bools {=bool} {=bool} [...]",
-            0b0110_1011, // compressed bools (dec value = 107)
-        ]
-    );
-}
-
-#[test]
-fn booleans_less_than_max_num_bool_flags() {
-    let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
-    write!(
-        g,
-        "encode 3 bools {=bool} {=bool} {=bool}",
-        false, true, true
-    );
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // "encode 3 bools {=bool} {=bool} {=bool}",
-            0b011, // compressed bools
-        ]
-    );
-}
-
-#[test]
-fn booleans_more_than_max_num_bool_flags() {
-    let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
-    write!(g, "encode 9 bools {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool} {=bool}",
-           false, true, true, false, true, false, true, true, false, true);
-    assert_eq!(
-        f.bytes(),
-        &[
-            index,       // "encode 8 bools {=bool} {=bool} {=bool} [...]",
-            0b0110_1011, // first 8 compressed bools
-            0b01,        // final compressed bools
-        ]
-    );
-}
-
-#[test]
-fn booleans_mixed() {
-    let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
-    write!(
-        g,
-        "encode mixed bools {=bool} {=bool} {=u8} {=bool}",
-        true, false, 42, true
-    );
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // "encode mixed bools {=bool} {=bool} {=u8} {=bool}",
-            42u8,  // intermediate `42`
-            0b101, // all compressed bools
-        ]
-    );
-}
-
-#[test]
-fn booleans_mixed_no_trailing_bool() {
-    let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
-    write!(g, "encode mixed bools {=bool} {=u8}", false, 42);
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // "encode mixed bools {=bool} {=u8}",
-            42u8, 0b0, // bool is put at the end of the args
-        ]
-    );
+    let g = defmt::export::make_formatter();
+    write!(g, "The answer is {=?}", 42u8);
+    check!([
+        inc(index, 1), // "The answer is {=?}"
+        inc(index, 2), // "{=u8}" / impl Format for u8
+        42u8,          // u8 value
+    ]);
 }
 
 #[test]
 fn bitfields_mixed() {
     let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
+    let g = defmt::export::make_formatter();
 
     write!(
         g,
         "bitfields {0=7..12}, {1=0..5}",
         0b1110_0101_1111_0000u16, 0b1111_0000u8
     );
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // bitfields {0=7..12}, {1=0..5}",
-            0b1111_0000,
-            0b1110_0101,   // u16
-            0b1111_0000u8, // u8
-        ]
-    );
+    check!([
+        index, // bitfields {0=7..12}, {1=0..5}",
+        0b1111_0000u8,
+        0b1110_0101u8, // u16
+        0b1111_0000u8, // u8
+    ]);
 }
 
 #[test]
 fn bitfields_across_octets() {
     let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
+    let g = defmt::export::make_formatter();
 
     write!(g, "bitfields {0=0..7} {0=9..14}", 0b0110_0011_1101_0010u16);
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // bitfields {0=0..7} {0=9..14}",
-            0b1101_0010,
-            0b0110_0011, // u16
-        ]
-    );
+    check!([
+        index, // bitfields {0=0..7} {0=9..14}",
+        0b1101_0010u8,
+        0b0110_0011u8, // u16
+    ]);
 }
 
 #[test]
 fn bitfields_truncate_lower() {
     let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
+    let g = defmt::export::make_formatter();
 
     write!(
         g,
         "bitfields {0=9..14}",
         0b0000_0000_0000_1111_0110_0011_1101_0010u32
     );
-    assert_eq!(
-        f.bytes(),
-        &[
-            index,       // bitfields {0=9..14}",
-            0b0110_0011, // the first octet should have been truncated away
-        ]
-    );
+    check!([
+        index,         // bitfields {0=9..14}",
+        0b0110_0011u8, // the first octet should have been truncated away
+    ]);
 }
 
 #[test]
 fn bitfields_assert_range_exclusive() {
     let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
+    let g = defmt::export::make_formatter();
 
     write!(g, "bitfields {0=6..8}", 0b1010_0101u8,);
-    assert_eq!(
-        f.bytes(),
-        &[
-            index, // "bitfields {0=6..8}"
-            0b1010_0101
-        ]
-    );
+    check!([
+        index, // "bitfields {0=6..8}"
+        0b1010_0101u8
+    ]);
 }
 
 #[test]
@@ -262,43 +160,16 @@ fn boolean_struct() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &X { y: false, z: true },
-        &[
+        [
             index, // "X {{ x: {=bool}, y: {=bool} }}"
-            0b01,  // y and z compressed together
+            0b0u8, // y
+            0b1u8, // z
         ],
     )
 }
 
-#[test]
-fn boolean_struct_mixed() {
-    #[derive(Format)]
-    struct X {
-        y: bool,
-        z: bool,
-    }
-
-    let index = fetch_string_index();
-    let f = &mut InternalFormatter::new();
-    let g = Formatter { inner: f };
-
-    write!(
-        g,
-        "mixed formats {=bool} {=?}",
-        true,
-        X { y: false, z: true }
-    );
-    assert_eq!(
-        f.bytes(),
-        &[
-            index,         // "mixed formats {=bool} {=?}",
-            inc(index, 1), // "X {{ x: {=bool}, y: {=bool} }}"
-            0b101,         // compressed struct bools
-        ]
-    );
-}
-
 #[test]
 fn single_struct() {
     #[derive(Format)]
@@ -308,13 +179,13 @@ fn single_struct() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &X { y: 1, z: 2 },
-        &[
+        [
             index, // "X {{ x: {=u8}, y: {=u16} }}"
-            1,     // x
-            2,     // y.low
-            0,     // y.high
+            1u8,   // x
+            2u8,   // y.low
+            0u8,   // y.high
         ],
     )
 }
@@ -329,18 +200,86 @@ fn single_struct_manual() {
 
     impl Format for X {
         fn format(&self, f: Formatter) {
-            defmt::write!(f, "X {{ x: {=u8}, y: {=u16} }}", self.y, self.z)
+            defmt::write!(f, "X {{ y: {=u8}, z: {=u16} }}", self.y, self.z)
         }
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &X { y: 1, z: 2 },
-        &[
-            index, // "X {{ x: {=u8}, y: {=u16} }}"
-            1,     // x
-            2,     // y.low
-            0,     // y.high
+        [
+            index,         // "{=__internal_FormatSequence}"
+            inc(index, 1), // "X {{ y: {=u8}, z: {=u16} }}"
+            1u8,           // y
+            2u16,          // z
+            0u16,          // terminator
+        ],
+    )
+}
+
+#[test]
+fn single_struct_manual_multiwrite() {
+    // Above `#[derive]`d impl should be equivalent to this:
+    struct X {
+        y: u8,
+        z: u16,
+    }
+
+    impl Format for X {
+        fn format(&self, f: Formatter) {
+            defmt::write!(f, "y={=u8}", self.y);
+            defmt::write!(f, "z={=u16}", self.z);
+        }
+    }
+
+    let index = fetch_string_index();
+    check_format!(
+        &X { y: 1, z: 2 },
+        [
+            index,         // "{=__internal_FormatSequence}"
+            inc(index, 1), // "y={=u8}"
+            1u8,           // y
+            inc(index, 2), // "z={=u16}"
+            2u16,          // z
+            0u16,          // terminator
+        ],
+    )
+}
+
+#[test]
+fn slice_struct_manual_multiwrite() {
+    // Above `#[derive]`d impl should be equivalent to this:
+    struct X {
+        y: u8,
+        z: u16,
+    }
+
+    impl Format for X {
+        fn format(&self, f: Formatter) {
+            defmt::write!(f, "y={=u8}", self.y);
+            defmt::write!(f, "z={=u16}", self.z);
+        }
+    }
+
+    let index = fetch_string_index();
+    check_format!(
+        &[X { y: 1, z: 2 }, X { y: 3, z: 4 }][..],
+        [
+            index,         // "{=[?]}"
+            2u32,          // len
+            inc(index, 1), // "{=__internal_FormatSequence}"
+            // first element
+            inc(index, 2), // "y={=u8}"
+            1u8,           // y
+            inc(index, 3), // "z={=u16}"
+            2u16,          // z
+            0u16,          // terminator
+            // second element
+            inc(index, 4), // "y={=u8}"
+            3u8,           // y
+            inc(index, 5), // "z={=u16}"
+            4u16,          // z
+            0u16,          // terminator
         ],
     )
 }
@@ -357,11 +296,11 @@ fn nested_struct() {
         z: u8,
     }
 
-    let val = 42;
+    let val = 42u8;
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &X { y: Y { z: val } },
-        &[
+        [
             index,         // "X {{ y: {=?} }}"
             inc(index, 1), // "Y {{ z: {=u8} }}"
             val,
@@ -375,12 +314,12 @@ fn tuple_struct() {
     struct Struct(u8, u16);
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Struct(0x1f, 0xaaaa),
-        &[
-            index, // "Struct({=u8}, {=u16})"
-            0x1f,  // u8
-            0xaa, 0xaa, // u16
+        [
+            index,     // "Struct({=u8}, {=u16})"
+            0x1fu8,    // u8
+            0xaaaau16, // u16
         ],
     );
 }
@@ -396,19 +335,19 @@ fn c_like_enum() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Enum::A,
-        &[
+        [
             index, //
-            0,     // `Enum::A`
+            0u8,   // `Enum::A`
         ],
     );
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Enum::B,
-        &[
+        [
             index, //
-            1,     // `Enum::B`
+            1u8,   // `Enum::B`
         ],
     );
 }
@@ -429,9 +368,9 @@ fn univariant_enum() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &NoData::Variant,
-        &[
+        [
             index, //
         ],
     );
@@ -442,12 +381,12 @@ fn univariant_enum() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Data::Variant(0x1f, 0xaaaa),
-        &[
-            index, //
-            0x1f,  // u8
-            0xaa, 0xaa, // u16
+        [
+            index,     //
+            0x1fu8,    // u8
+            0xaaaau16, // u16
         ],
     );
 }
@@ -476,45 +415,45 @@ fn nested_enum() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Outer::Variant1 {
             pre: 0xEE,
             inner: Inner::A(CLike::B, 0x07),
             post: 0xAB,
         },
-        &[
+        [
             index,         //
-            0,             // `Outer::Variant1`
-            0xEE,          // u8 pre
+            0u8,           // `Outer::Variant1`
+            0xEEu8,        // u8 pre
             inc(index, 1), // `Inner`'s formatting string
-            0,             // `Inner::A`
+            0u8,           // `Inner::A`
             inc(index, 2), // `CLike`'s formatting string
-            1,             // `CLike::B`
-            0x07,          // u8
-            0xAB,          // u8 post
+            1u8,           // `CLike::B`
+            0x07u8,        // u8
+            0xABu8,        // u8 post
         ],
     );
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Outer::Variant2,
-        &[
+        [
             index, //
-            1,     // `Outer::Variant2`
+            1u8,   // `Outer::Variant2`
         ],
     );
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &Outer::Variant3(Inner::A(CLike::B, 0x07)),
-        &[
+        [
             index,         //
-            2,             // `Outer::Variant3`
+            2u8,           // `Outer::Variant3`
             inc(index, 1), // `Inner`'s formatting string
-            0,             // `Inner::A`
+            0u8,           // `Inner::A`
             inc(index, 2), // `CLike`'s formatting string
-            1,             // `CLike::B`
-            0x07,          // u8
+            1u8,           // `CLike::B`
+            0x07u8,        // u8
         ],
     );
 }
@@ -523,14 +462,14 @@ fn nested_enum() {
 fn slice() {
     let index = fetch_string_index();
     let val: &[u8] = &[23u8, 42u8];
-    check_format_implementation(
+    check_format!(
         val,
-        &[
-            index,           // "{=[?]}"
-            val.len() as u8, // length
-            inc(index, 1),   // "{=u8}"
-            23,              // val[0]
-            42,              // val[1]
+        [
+            index,            // "{=[?]}"
+            val.len() as u32, // length
+            inc(index, 1),    // "{=u8}"
+            23u8,             // val[0]
+            42u8,             // val[1]
         ],
     )
 }
@@ -539,14 +478,14 @@ fn slice() {
 fn slice_of_usize() {
     let index = fetch_string_index();
     let val: &[usize] = &[23usize, 42];
-    check_format_implementation(
+    check_format!(
         val,
-        &[
-            index,           // "{=[?]}"
-            val.len() as u8, // length
-            inc(index, 1),   // "{=usize}"
-            23,              // val[0]
-            42,              // val[1]
+        [
+            index,            // "{=[?]}"
+            val.len() as u32, // length
+            inc(index, 1),    // "{=usize}"
+            23u32,            // val[0]
+            42u32,            // val[1]
         ],
     )
 }
@@ -555,13 +494,15 @@ fn slice_of_usize() {
 fn slice_of_bools() {
     let index = fetch_string_index();
     let val: &[bool] = &[true, true, false];
-    check_format_implementation(
+    check_format!(
         val,
-        &[
-            index,           // "{=[?]}"
-            val.len() as u8, // length
-            inc(index, 1),   // "{=bool}"
-            0b110,           // compressed bools: true, true, false
+        [
+            index,            // "{=[?]}"
+            val.len() as u32, // length
+            inc(index, 1),    // "{=bool}"
+            0b1u8,
+            0b1u8,
+            0b0u8,
         ],
     )
 }
@@ -569,144 +510,122 @@ fn slice_of_bools() {
 #[test]
 fn format_primitives() {
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &42u8,
-        &[
+        [
             index, // "{=u8}"
-            42,
+            42u8,
         ],
     );
-    check_format_implementation(
+    check_format!(
         &42u16,
-        &[
+        [
             inc(index, 1), // "{=u16}"
-            42,
-            0,
+            42u16,
         ],
     );
-    check_format_implementation(
+    check_format!(
         &513u16,
-        &[
+        [
             inc(index, 2), // "{=u16}"
-            1,
-            2,
+            513u16,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &42u32,
-        &[
+        [
             inc(index, 3), // "{=u32}"
-            42,
-            0,
-            0,
-            0,
+            42u32,
         ],
     );
-    check_format_implementation(
+    check_format!(
         &513u32,
-        &[
+        [
             inc(index, 4), // "{=u32}"
-            1,
-            2,
-            0,
-            0,
+            513u32,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &5.13f32,
-        &[
+        [
             inc(index, 5), // "{=f32}"
-            246,
-            40,
-            164,
-            64,
+            246u8,
+            40u8,
+            164u8,
+            64u8,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &42i8,
-        &[
+        [
             inc(index, 6), // "{=i8}"
-            42,
+            42u8,
         ],
     );
-    check_format_implementation(
+    check_format!(
         &-42i8,
-        &[
+        [
             inc(index, 7), // "{=i8}"
             -42i8 as u8,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &None::<u8>,
-        &[
+        [
             inc(index, 8), // "<option-format-string>"
-            0,             // None discriminant
+            0u8,           // None discriminant
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &Some(42u8),
-        &[
+        [
             inc(index, 9),  // "<option-format-string>"
-            1,              // Some discriminant
+            1u8,            // Some discriminant
             inc(index, 10), // "{=u8}"
-            42,             // Some.0 field
+            42u8,           // Some.0 field
         ],
     );
 
-    check_format_implementation(&-1isize, &[inc(index, 11), 0b0000_0001]);
-    check_format_implementation(&-128isize, &[inc(index, 12), 0xff, 0b0000_0001]);
+    check_format!(&-1isize, [inc(index, 11), -1i32]);
+    check_format!(&-128isize, [inc(index, 12), -128i32]);
 
-    check_format_implementation(
+    check_format!(
         &true,
-        &[
+        [
             inc(index, 13), // "{=bool}"
-            0b1,
+            0b1u8,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &513u64,
-        &[
+        [
             inc(index, 14), // "{=u64}"
-            1,
-            2,
-            0,
-            0,
-            0,
-            0,
-            0,
-            0,
+            513u64,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &-2i64,
-        &[
+        [
             inc(index, 15), // "{=i64}"
-            0xFE,
-            0xFF,
-            0xFF,
-            0xFF,
-            0xFF,
-            0xFF,
-            0xFF,
-            0xFF,
+            -2i64,
         ],
     );
 
-    check_format_implementation(
+    check_format!(
         &'a',
-        &[
+        [
             inc(index, 16), // "{=char}"
-            0x61,
-            0x00,
-            0x00,
-            0x00,
+            0x61u8,
+            0x00u8,
+            0x00u8,
+            0x00u8,
         ],
     );
 }
@@ -715,9 +634,9 @@ fn format_primitives() {
 fn istr() {
     let index = fetch_string_index();
     let interned = defmt::intern!("interned string contents");
-    check_format_implementation(
+    check_format!(
         &interned,
-        &[
+        [
             inc(index, 1), // "{=istr}"
             index,
         ],
@@ -728,21 +647,24 @@ fn istr() {
 fn format_arrays() {
     let index = fetch_string_index();
     let array: [u16; 0] = [];
-    check_format_implementation(&array, &[index]);
+    check_format!(
+        &array,
+        [
+            index,         // "{=[?;0]}"
+            inc(index, 1), // "{=u16}"
+        ]
+    );
 
     let index = fetch_string_index();
     let array: [u16; 3] = [1, 256, 257];
-    check_format_implementation(
+    check_format!(
         &array,
-        &[
+        [
             index,         // "{=[?;3]}"
             inc(index, 1), // "{=u16}"
-            1,             // [0].low
-            0,             // [0].high
-            0,             // [1].low
-            1,             // [1].high
-            1,             // [2].low
-            1,             // [2].high
+            1u16,          // [0]
+            256u16,        // [1]
+            257u16,        // [2]
         ],
     );
 }
@@ -751,18 +673,15 @@ fn format_arrays() {
 fn format_slice_of_primitives() {
     let index = fetch_string_index();
     let slice: &[u16] = &[1, 256, 257];
-    check_format_implementation(
+    check_format!(
         slice,
-        &[
-            index,             // "{=[?]}"
-            slice.len() as u8, //
-            inc(index, 1),     // "{=u16}"
-            1,                 // [0].low
-            0,                 // [0].high
-            0,                 // [1].low
-            1,                 // [1].high
-            1,                 // [2].low
-            1,                 // [2].high
+        [
+            index,              // "{=[?]}"
+            slice.len() as u32, //
+            inc(index, 1),      // "{=u16}"
+            1u16,               // [0]
+            256u16,             // [1]
+            257u16,             // [2]
         ],
     );
 }
@@ -781,17 +700,18 @@ fn format_slice_of_structs() {
 
     let index = fetch_string_index();
     let slice: &[_] = &[X { y: Y { z: 42 } }, X { y: Y { z: 24 } }];
-    check_format_implementation(
+    check_format!(
         slice,
-        &[
-            index,             // "{=[?]}"
-            slice.len() as u8, //
+        [
+            index,              // "{=[?]}"
+            slice.len() as u32, //
             // first element
             inc(index, 1), // "X {{ y: {=?} }}"
             inc(index, 2), // "Y {{ z: {=u8} }}"
-            42,            // [0].y.z
-            // second element: no tags
-            24, // [1].y.z
+            42u8,          // [0].y.z
+            // second element: no outer tag
+            inc(index, 3), // "Y {{ z: {=u8} }}"
+            24u8,          // [1].y.z
         ],
     );
 }
@@ -800,29 +720,23 @@ fn format_slice_of_structs() {
 fn format_slice_of_slices() {
     let index = fetch_string_index();
     let slice: &[&[u16]] = &[&[256, 257], &[258, 259, 260]];
-    check_format_implementation(
+    check_format!(
         slice,
-        &[
-            index,             // "{=[?]}"
-            slice.len() as u8, //
+        [
+            index,              // "{=[?]}"
+            slice.len() as u32, //
+            inc(index, 1),      // "{=[?]}"
             // first slice
-            inc(index, 1), // "{=[?]}"
-            slice[0].len() as u8,
-            // its first element
+            slice[0].len() as u32,
             inc(index, 2), // "{=u16}"
-            0,             // [0][0].low
-            1,             // [0][0].high
-            // its second element: no tag
-            1, // [0][1].low
-            1, // [0][1].high
-            // second slice: no tags
-            slice[1].len() as u8,
-            2, // [1][0].low
-            1, // [1][0].high
-            3, // [1][1].low
-            1, // [1][1].high
-            4, // [1][2].low
-            1, // [1][2].high
+            256u16,        // [0][0]
+            257u16,        // [0][1]
+            // second slice
+            slice[1].len() as u32,
+            inc(index, 3), // "{=u16}"
+            258u16,        // [1][0]
+            259u16,        // [1][1]
+            260u16,        // [1][2]
         ],
     );
 }
@@ -831,23 +745,23 @@ fn format_slice_of_slices() {
 fn format_slice_enum_slice() {
     let index = fetch_string_index();
     let slice: &[Option<&[u8]>] = &[None, Some(&[42, 43])];
-    check_format_implementation(
+    check_format!(
         slice,
-        &[
-            index,             // "{=[?]}"
-            slice.len() as u8, //
+        [
+            index,              // "{=[?]}"
+            slice.len() as u32, //
             // first optional slice
             inc(index, 1), // "None|Some({=?})"
-            0,             // discriminant
+            0u8,           // discriminant
             // second optional slice
             // omitted: "None|Some({=?})" index
-            1,             // discriminant
+            1u8,           // discriminant
             inc(index, 2), // "{=[?]}" (the ? behind "Some({=?})")
-            2,             // length of second optional slice
+            2u32,          // length of second optional slice
             inc(index, 3), // "{=u8}" (the ? behind "{=[?]}")
-            42,
+            42u8,
             // omitted: "{=u8}" index
-            43,
+            43u8,
         ],
     );
 }
@@ -862,21 +776,21 @@ fn format_slice_enum_generic_struct() {
 
     let index = fetch_string_index();
     let slice: &[Option<S<u8>>] = &[None, Some(S { x: 42, y: 43 })];
-    check_format_implementation(
+    check_format!(
         slice,
-        &[
-            index,             // "{=[?]}"
-            slice.len() as u8, //
+        [
+            index,              // "{=[?]}"
+            slice.len() as u32, //
             // first optional element
             inc(index, 1), // "None|Some({=?})"
-            0,             // discriminant
+            0u8,           // discriminant
             // second optional element
             // omitted: "None|Some({=?})" index
-            1,             // discriminant
+            1u8,           // discriminant
             inc(index, 2), // "S {{ x: {=u8}, y: {=?} }}" (the ? behind "Some({=?})")
-            42,            // S.x
+            42u8,          // S.x
             inc(index, 3), // "{=u8}" (the ? behind S.y)
-            43,            // S. y
+            43u8,          // S. y
         ],
     );
 }
@@ -895,24 +809,21 @@ fn derive_with_bounds() {
     }
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &S { val: 0 },
-        &[
+        [
             index,         // "S {{ val: {=?} }}"
             inc(index, 1), // "{=i32}"
-            0,
-            0,
-            0,
-            0,
+            0u32,
         ],
     );
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &S2 { a: &1, b: &2 },
-        &[
+        [
             index, // "S2 { a: {=u8}, b: {=u8} }}"
-            1, 2,
+            1u8, 2u8,
         ],
     );
 }
@@ -926,40 +837,14 @@ fn format_bools() {
     struct B(bool);
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &(A(true), B(true)),
-        &[
+        [
             index,         // "({=?}, {=?})"
             inc(index, 1), // "A({=bool})"
+            0b1u8,         // A
             inc(index, 2), // "B({=bool})"
-            0b11,          // compressed bools
-        ],
-    );
-}
-
-#[test]
-fn issue_208() {
-    #[derive(Format)]
-    struct DhcpReprMin {
-        pub broadcast: bool,
-        pub a: [u8; 2],
-    }
-
-    let dhcp_repr = DhcpReprMin {
-        broadcast: true,
-        a: [10, 10],
-    };
-
-    let index = fetch_string_index();
-    check_format_implementation(
-        &dhcp_repr,
-        &[
-            index,         // "DhcpReprMin {{ broadcast: {=bool}, a: {=?} }}"
-            inc(index, 1), // "{=[?;2]}"
-            inc(index, 2), // "{=u8}"
-            10,            // a[0]
-            10,            // a[1]
-            1,             // compressed bools
+            0b1u8,         // B
         ],
     );
 }
@@ -998,12 +883,15 @@ fn enum_variants() {
     let e = EnumSmall::A2;
 
     let index = fetch_string_index();
-    check_format_implementation(&e, &[index, 2]);
+    check_format!(&e, [index, 2u8]);
+
+    let e = EnumLarge::A002;
+    let index = fetch_string_index();
+    check_format!(&e, [index, 2u16]);
 
     let e = EnumLarge::A269;
-
     let index = fetch_string_index();
-    check_format_implementation(&e, &[index, 13, 1]);
+    check_format!(&e, [index, 269u16]);
 }
 
 #[test]
@@ -1016,14 +904,14 @@ fn derive_str() {
     let s = S { x: "hi" };
 
     let index = fetch_string_index();
-    check_format_implementation(
+    check_format!(
         &s,
-        &[
+        [
             index, // "S {{ s: {:str} }}" (NOTE: `s` field is not {:?})
             // so no extra format string index here
-            2,   // s.x.len()
-            104, // b'h'
-            105, // b'i'
+            2u32,  // s.x.len()
+            104u8, // b'h'
+            105u8, // b'i'
         ],
     );
 }
@@ -1031,7 +919,7 @@ fn derive_str() {
 #[test]
 fn core_fmt_adapters() {
     let index = fetch_string_index();
-    check_format_implementation(&Debug2Format(&123u8), &[index, b'1', b'2', b'3', 0xff]);
+    check_format!(&Debug2Format(&123u8), [index, b'1', b'2', b'3', 0xffu8]);
     let index = fetch_string_index();
-    check_format_implementation(&Display2Format(&123u8), &[index, b'1', b'2', b'3', 0xff]);
+    check_format!(&Display2Format(&123u8), [index, b'1', b'2', b'3', 0xffu8]);
 }
